#include <R.h>
#include <Rinternals.h>
#include <R_ext/Lapack.h>
#include <R_ext/BLAS.h>

#ifdef _OPENMP
#include <omp.h>
#endif

extern "C" {

  SEXP dMvN(SEXP n_r, SEXP y_r, SEXP mu_r, SEXP C_r, SEXP nThreads_r){
    
    int n = INTEGER(n_r)[0];
    double *y = REAL(y_r);
    double *mu = REAL(mu_r);
    double *C = REAL(C_r);
    int nThreads = INTEGER(nThreads_r)[0];
    
#ifdef _OPENMP
    omp_set_num_threads(nThreads);
#else
    if(nThreads > 1){
      warning("n.omp.threads = %i requested however source code was not compiled with OpenMP support.", nThreads);
      nThreads = 1;
    }
#endif
    
    //common variables
    int i, info, nProtect = 0;
    char const *lower = "L";
    char const *nUnit = "N";
    char const *ntran = "N";

    const int incOne = 1;
    const double pi = 3.14159265359;

    double det = 0;
    double Q;
    double *z = (double *) R_alloc(n, sizeof(double)); 
    SEXP result = PROTECT(allocVector(REALSXP, 1)); nProtect++;
 
   //z = mu-y
   for(i = 0; i < n; i++){
     z[i] = mu[i]-y[i];
   }
 
   //Cholesky factorization (note C is overwritten, so you might want to make a copy if you'll need it later)
   F77_NAME(dpotrf)(lower, &n, C, &n, &info); if(info != 0){error("c++ error: dpotrf failed\n");}

   //log determinant
   for(i = 0; i < n; i++){
     det += 2.0*log(C[i*n+i]);
   }
   
   //solve Lu = z for u (note u is now stored in z, just to be confusing)
   F77_NAME(dtrsv)(lower, ntran, nUnit, &n, C, &n, z, &incOne);
   
   Q = F77_NAME(ddot)(&n, z, &incOne, z, &incOne);
   
   REAL(result)[0] = -0.5*n*log(2.0*pi)-0.5*det-0.5*Q;

   UNPROTECT(nProtect);

   return result;
 }
}
